import torch
import torch.nn as nn


class Model(nn.Module):
    """
    Model that performs a matrix multiplication, division, summation, and scaling.
    """

    def __init__(self, input_size, hidden_size, scaling_factor):
        super(Model, self).__init__()
        self.weight = nn.Parameter(torch.randn(hidden_size, input_size))
        self.scaling_factor = scaling_factor

    def forward(self, x):
        """
        Args:
            x (torch.Tensor): Input tensor of shape (batch_size, input_size).
        Returns:
            torch.Tensor: Output tensor of shape (batch_size, hidden_size).
        """
        x = torch.matmul(x, self.weight.T)
        x = x / 2
        x = torch.sum(x, dim=1, keepdim=True)
        x = x * self.scaling_factor
        return x



import mlx.core as mx  # noqa
import mlx.nn as nn
from typing import Union


class Model(nn.Module):
    """
    Model that performs a 2D convolution.

    Args:
        in_channels (int): Number of input channels.
        out_channels (int): Number of output channels.
        kernel_size (Union[int, tuple]): Size of the convolution kernel.
        stride (Union[int, tuple]): Stride of the convolution. Default is 1.
    """

    def __init__(self, in_channels: int, out_channels: int, kernel_size: Union[int, tuple], stride: Union[int, tuple] = 1):
        super(Model, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size, stride)

    def __call__(self, x):
        """
        Args:
            x (mx.array): Input tensor of shape (batch_size, height, width, in_channels).
        Returns:
            mx.array: Output tensor of shape (batch_size, height, width, out_channels).
        """
        return self.conv(x)



import argparse
import sys
import pathlib
import math
from rich.console import Console
from rich.live import Live
from rich.panel import Panel
from rich.traceback import install
from .api import start_optimization_session, evaluate_feedback_then_suggest_next_solution, get_optimization_session_status
from .panels import (
    SummaryPanel,
    PlanPanel,
    Node,
    MetricTreePanel,
    EvaluationOutputPanel,
    SolutionPanels,
    create_optimization_layout,
    create_end_optimization_layout,
)
from .utils import (
    read_api_keys_from_env,
    read_additional_instructions,
    read_from_path,
    write_to_path,
    run_evaluation,
    smooth_update,
    format_number,
)

install(show_locals=True)
console = Console()


def main() -> None:
    """Main function for the Weco CLI."""
    parser = argparse.ArgumentParser(
        description="[bold cyan]Weco CLI[/]", formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument("--source", type=str, required=True, help="Path to the Python source code (e.g. optimize.py)")
    parser.add_argument(
        "--eval-command", type=str, required=True, help="Command to run for evaluation (e.g. 'python eval.py --arg1=val1')"
    )
    parser.add_argument("--metric", type=str, required=True, help="Metric to optimize")
    parser.add_argument(
        "--maximize",
        type=str,
        choices=["true", "false"],
        required=True,
        help="Specify 'true' to maximize the metric or 'false' to minimize.",
    )
    parser.add_argument("--steps", type=int, required=True, help="Number of steps to run")
    parser.add_argument("--model", type=str, required=True, help="Model to use for optimization")
    parser.add_argument(
        "--additional-instructions",
        default=None,
        type=str,
        help="Description of additional instruction or path to a file containing additional instructions",
    )
    args = parser.parse_args()

    try:
        with console.status("[bold green]Loading Modules..."):
            # Define optimization session config
            evaluation_command = args.eval_command
            metric_name = args.metric
            maximize = args.maximize == "true"
            steps = args.steps
            code_generator_config = {"model": args.model}
            evaluator_config = {"model": args.model}
            search_policy_config = {
                "num_drafts": max(1, math.ceil(0.15 * steps)),  # 15% of steps
                "debug_prob": 0.5,
                "max_debug_depth": max(1, math.ceil(0.1 * steps)),  # 10% of steps
            }
            # Read additional instructions
            additional_instructions = read_additional_instructions(additional_instructions=args.additional_instructions)
            # Read source code
            source_fp = pathlib.Path(args.source)
            source_code = read_from_path(fp=source_fp, is_json=False)
            # Read API keys
            api_keys = read_api_keys_from_env()
            # API request timeout
            timeout = 800

        # Initialize panels
        summary_panel = SummaryPanel(maximize=maximize, metric_name=metric_name, total_steps=steps, model=args.model)
        plan_panel = PlanPanel()
        solution_panels = SolutionPanels(metric_name=metric_name)
        eval_output_panel = EvaluationOutputPanel()
        tree_panel = MetricTreePanel(maximize=maximize)
        layout = create_optimization_layout()
        end_optimization_layout = create_end_optimization_layout()

        # Start optimization session
        session_response = start_optimization_session(
            console=console,
            source_code=source_code,
            evaluation_command=evaluation_command,
            metric_name=metric_name,
            maximize=maximize,
            steps=steps,
            code_generator_config=code_generator_config,
            evaluator_config=evaluator_config,
            search_policy_config=search_policy_config,
            additional_instructions=additional_instructions,
            api_keys=api_keys,
            timeout=timeout,
        )

        # Define the refresh rate
        refresh_rate = 4
        with Live(layout, refresh_per_second=refresh_rate, screen=True) as live:
            # Define the runs directory (.runs/<session-id>)
            session_id = session_response["session_id"]
            runs_dir = pathlib.Path(".runs") / session_id
            runs_dir.mkdir(parents=True, exist_ok=True)

            # Save the original code (.runs/<session-id>/original.py)
            runs_copy_source_fp = runs_dir / "original.py"
            write_to_path(fp=runs_copy_source_fp, content=source_code)

            # Write the code string to the source file path
            # Do this after the original code is saved
            write_to_path(fp=source_fp, content=session_response["code"])

            # Update the panels with the initial solution
            # Add session id now that we have it
            summary_panel.session_id = session_id
            # Set the step of the progress bar
            summary_panel.set_step(step=0)
            # Update the token counts
            summary_panel.update_token_counts(usage=session_response["usage"])
            # Update the plan
            plan_panel.update(plan=session_response["plan"])
            # Build the metric tree
            tree_panel.build_metric_tree(
                nodes=[
                    {
                        "solution_id": session_response["solution_id"],
                        "parent_id": None,
                        "code": session_response["code"],
                        "step": 0,
                        "metric_value": None,
                        "is_buggy": False,
                    }
                ]
            )
            # Set the current solution as unevaluated since we haven't run the evaluation function and fed it back to the model yet
            tree_panel.set_unevaluated_node(node_id=session_response["solution_id"])
            # Update the solution panels with the initial solution and get the panel displays
            solution_panels.update(
                current_node=Node(
                    id=session_response["solution_id"],
                    parent_id=None,
                    code=session_response["code"],
                    metric=None,
                    is_buggy=False,
                ),
                best_node=None,
            )
            current_solution_panel, best_solution_panel = solution_panels.get_display(current_step=0)

            # Update the entire layout
            smooth_update(
                live=live,
                layout=layout,
                sections_to_update=[
                    ("summary", summary_panel.get_display()),
                    ("plan", plan_panel.get_display()),
                    ("tree", tree_panel.get_display()),
                    ("current_solution", current_solution_panel),
                    ("best_solution", best_solution_panel),
                    ("eval_output", eval_output_panel.get_display()),
                ],
                transition_delay=0.1,  # Slightly longer delay for initial display
            )

            # Run evaluation on the initial solution
            term_out = run_evaluation(eval_command=args.eval_command)

            # Update the evaluation output panel
            eval_output_panel.update(output=term_out)
            smooth_update(
                live=live,
                layout=layout,
                sections_to_update=[("eval_output", eval_output_panel.get_display())],
                transition_delay=0.1,
            )

            for step in range(1, steps):
                # Evaluate the current output and get the next solution
                eval_and_next_solution_response = evaluate_feedback_then_suggest_next_solution(
                    console=console,
                    session_id=session_id,
                    execution_output=term_out,
                    additional_instructions=additional_instructions,
                    api_keys=api_keys,
                    timeout=timeout,
                )
                # Save next solution (.runs/<session-id>/step_<step>.py)
                write_to_path(fp=runs_dir / f"step_{step}.py", content=eval_and_next_solution_response["code"])

                # Write the next solution to the source file
                write_to_path(fp=source_fp, content=eval_and_next_solution_response["code"])

                # Get the optimization session status for
                # the best solution, its score, and the history to plot the tree
                status_response = get_optimization_session_status(
                    console=console, session_id=session_id, include_history=True, timeout=timeout
                )

                # Update the step of the progress bar
                summary_panel.set_step(step=step)
                # Update the token counts
                summary_panel.update_token_counts(usage=eval_and_next_solution_response["usage"])
                # Update the plan
                plan_panel.update(plan=eval_and_next_solution_response["plan"])
                # Build the metric tree
                tree_panel.build_metric_tree(nodes=status_response["history"])
                # Set the current solution as unevaluated since we haven't run the evaluation function and fed it back to the model yet
                tree_panel.set_unevaluated_node(node_id=eval_and_next_solution_response["solution_id"])

                # Update the solution panels with the next solution and best solution (and score)
                # Figure out if we have a best solution so far
                if status_response["best_result"] is not None:
                    best_solution_node = Node(
                        id=status_response["best_result"]["solution_id"],
                        parent_id=status_response["best_result"]["parent_id"],
                        code=status_response["best_result"]["code"],
                        metric=status_response["best_result"]["metric_value"],
                        is_buggy=status_response["best_result"]["is_buggy"],
                    )
                else:
                    best_solution_node = None

                # Create a node for the current solution
                current_solution_node = None
                for node in status_response["history"]:
                    if node["solution_id"] == eval_and_next_solution_response["solution_id"]:
                        current_solution_node = Node(
                            id=node["solution_id"],
                            parent_id=node["parent_id"],
                            code=node["code"],
                            metric=node["metric_value"],
                            is_buggy=node["is_buggy"],
                        )
                if current_solution_node is None:
                    raise ValueError("Current solution node not found in history")
                # Update the solution panels with the current and best solution
                solution_panels.update(current_node=current_solution_node, best_node=best_solution_node)
                current_solution_panel, best_solution_panel = solution_panels.get_display(current_step=step)

                # Clear evaluation output since we are running a evaluation on a new solution
                eval_output_panel.clear()

                # Update displays with smooth transitions
                smooth_update(
                    live=live,
                    layout=layout,
                    sections_to_update=[
                        ("summary", summary_panel.get_display()),
                        ("plan", plan_panel.get_display()),
                        ("tree", tree_panel.get_display()),
                        ("current_solution", current_solution_panel),
                        ("best_solution", best_solution_panel),
                        ("eval_output", eval_output_panel.get_display()),
                    ],
                    transition_delay=0.08,  # Slightly longer delay for more noticeable transitions
                )

                # Run evaluation on the current solution
                term_out = run_evaluation(eval_command=args.eval_command)
                eval_output_panel.update(output=term_out)

                # Update evaluation output with a smooth transition
                smooth_update(
                    live=live,
                    layout=layout,
                    sections_to_update=[("eval_output", eval_output_panel.get_display())],
                    transition_delay=0.1,  # Slightly longer delay for evaluation results
                )

            # Ensure we pass evaluation results for the last step's generated solution
            eval_and_next_solution_response = evaluate_feedback_then_suggest_next_solution(
                console=console,
                session_id=session_id,
                execution_output=term_out,
                additional_instructions=additional_instructions,
                api_keys=api_keys,
                timeout=timeout,
            )

            # Update the progress bar
            summary_panel.set_step(step=steps)
            # Update the token counts
            summary_panel.update_token_counts(usage=eval_and_next_solution_response["usage"])
            # No need to update the plan panel since we have finished the optimization
            # Get the optimization session status for
            # the best solution, its score, and the history to plot the tree
            status_response = get_optimization_session_status(
                console=console, session_id=session_id, include_history=True, timeout=timeout
            )
            # Build the metric tree
            tree_panel.build_metric_tree(nodes=status_response["history"])
            # No need to set any solution to unevaluated since we have finished the optimization
            # and all solutions have been evaluated
            # No neeed to update the current solution panel since we have finished the optimization
            # We only need to update the best solution panel
            # Figure out if we have a best solution so far
            if status_response["best_result"] is not None:
                best_solution_node = Node(
                    id=status_response["best_result"]["solution_id"],
                    parent_id=status_response["best_result"]["parent_id"],
                    code=status_response["best_result"]["code"],
                    metric=status_response["best_result"]["metric_value"],
                    is_buggy=status_response["best_result"]["is_buggy"],
                )
            else:
                best_solution_node = None
            solution_panels.update(current_node=None, best_node=best_solution_node)
            _, best_solution_panel = solution_panels.get_display(current_step=steps)

            # Update the end optimization layout
            final_message = (
                f"{summary_panel.metric_name.capitalize()} {'maximized' if summary_panel.maximize else 'minimized'}! Best solution {summary_panel.metric_name.lower()} = [green]{status_response['best_result']['metric_value']}[/] 🏆"
                if best_solution_node is not None
                else "[red] No solution found.[/]"
            )
            end_optimization_layout["summary"].update(summary_panel.get_display(final_message=final_message))
            end_optimization_layout["tree"].update(tree_panel.get_display())
            end_optimization_layout["best_solution"].update(best_solution_panel)

            # Save optimization results
            # If the best solution does not exist or is has not been measured at the end of the optimization
            # save the original solution as the best solution
            best_solution_code = best_solution_node.code
            best_solution_score = best_solution_node.metric
            if best_solution_code is None or best_solution_score is None:
                best_solution_content = (
                    f"# Weco could not find a better solution\n\n{read_from_path(fp=runs_copy_source_fp, is_json=False)}"
                )
            else:
                # Format score for the comment
                best_score_str = (
                    format_number(best_solution_score)
                    if best_solution_score is not None and isinstance(best_solution_score, (int, float))
                    else "N/A"
                )
                best_solution_content = f"# Best solution from Weco with a score of {best_score_str}\n\n{best_solution_code}"

            # Save best solution to .runs/<session-id>/best.py
            write_to_path(fp=runs_dir / "best.py", content=best_solution_content)

            # write the best solution to the source file
            write_to_path(fp=source_fp, content=best_solution_content)

        console.print(end_optimization_layout)

    except Exception as e:
        console.print(Panel(f"[bold red]Error: {str(e)}", title="[bold red]Error", border_style="red"))
        sys.exit(1)



import time
import sys
import pathlib
import importlib
import traceback
import torch
import torch.nn as nn
import torch.nn.functional as F
import math


########################################################
# Baseline
########################################################
class Model(nn.Module):
    """
    A vanilla multi-head masked self-attention layer with a projection at the end.
    """

    def __init__(self, n_embd, n_head, attn_pdrop, resid_pdrop, max_seqlen):
        super().__init__()
        assert n_embd % n_head == 0
        # key, query, value projections for all heads, but in a batch
        self.c_attn = nn.Linear(n_embd, 3 * n_embd)
        # output projection
        self.c_proj = nn.Linear(n_embd, n_embd)
        # regularization
        self.attn_dropout = nn.Dropout(attn_pdrop)
        self.resid_dropout = nn.Dropout(resid_pdrop)
        # causal mask to ensure that attention is only applied to the left in the input sequence
        self.register_buffer("bias", torch.tril(torch.ones(max_seqlen, max_seqlen)).view(1, 1, max_seqlen, max_seqlen))
        self.n_head = n_head
        self.n_embd = n_embd

    def forward(self, x):
        B, T, C = x.size()  # batch size, sequence length, embedding dimensionality (n_embd)
        # calculate query, key, values for all heads in batch and move head forward to be the batch dim
        q, k, v = self.c_attn(x).split(self.n_embd, dim=2)
        k = k.view(B, T, self.n_head, C // self.n_head).transpose(1, 2)  # (B, nh, T, hs)
        q = q.view(B, T, self.n_head, C // self.n_head).transpose(1, 2)  # (B, nh, T, hs)
        v = v.view(B, T, self.n_head, C // self.n_head).transpose(1, 2)  # (B, nh, T, hs)

        # causal self-attention; Self-attend: (B, nh, T, hs) x (B, nh, hs, T) -> (B, nh, T, T)
        att = (q @ k.transpose(-2, -1)) * (1.0 / math.sqrt(k.size(-1)))
        att = att.masked_fill(self.bias[:, :, :T, :T] == 0, float("-inf"))
        att = F.softmax(att, dim=-1)
        att = self.attn_dropout(att)
        y = att @ v  # (B, nh, T, T) x (B, nh, T, hs) -> (B, nh, T, hs)
        y = y.transpose(1, 2).contiguous().view(B, T, C)  # re-assemble all head outputs side by side
        # output projection
        y = self.resid_dropout(self.c_proj(y))
        return y


########################################################
# Weco Solution
########################################################
def load_module_from_path(module_path: str, add_to_sys_modules: bool = False):
    # Clean out all old compiled extensions to prevent namespace collisions during build
    module_path = pathlib.Path(module_path)
    name = module_path.stem
    spec = importlib.util.spec_from_file_location(name, module_path)
    mod = importlib.util.module_from_spec(spec)  # type: ignore
    if add_to_sys_modules:
        sys.modules[name] = mod
    spec.loader.exec_module(mod)  # type: ignore
    return mod


########################################################
# Benchmark
########################################################
def get_inputs(batch_size, seq_len, n_embd, device):
    return torch.randn(batch_size, seq_len, n_embd, device=device, dtype=torch.float32)


@torch.no_grad()
def bench(f, inputs, n_warmup, n_rep):
    # warmup
    for _ in range(n_warmup):
        f(inputs)  # noqa

    # benchmark
    t_avg = 0.0
    for _ in range(n_rep):
        torch.cuda.empty_cache()  # Clear cache before timing
        start_time = time.time()
        f(inputs)
        torch.cuda.synchronize()  # Wait for all computations to complete
        t_avg += time.time() - start_time
    t_avg /= n_rep * 1e-3
    return t_avg


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--solution-path", type=str, required=True)
    args = parser.parse_args()

    # benchmarking parameters
    n_correctness_trials = 10
    n_warmup = 1000
    n_rep = 5000

    # init parameters
    max_seqlen = 512
    seq_len = 256
    n_embd = 768
    n_head = 8
    # turn off dropout to measure correctness well
    attn_pdrop = 0.0
    resid_pdrop = 0.0

    # input parameters
    batch_size = 32

    # load solution module
    try:
        torch.manual_seed(0)
        solution_module = load_module_from_path(args.solution_path, add_to_sys_modules=False)
        solution_model = solution_module.Model(
            n_embd=n_embd, n_head=n_head, attn_pdrop=attn_pdrop, resid_pdrop=resid_pdrop, max_seqlen=max_seqlen
        ).to("cuda")
        assert isinstance(solution_model, nn.Module)
    except Exception:
        print(f"Candidate module initialization failed: {traceback.format_exc()}")
        exit(1)

    torch.manual_seed(0)
    baseline_model = Model(
        n_embd=n_embd, n_head=n_head, attn_pdrop=attn_pdrop, resid_pdrop=resid_pdrop, max_seqlen=max_seqlen
    ).to("cuda")

    # measure correctness
    max_diff_avg = 0
    for _ in range(n_correctness_trials):
        inputs = get_inputs(batch_size=batch_size, seq_len=seq_len, n_embd=n_embd, device="cuda")
        with torch.no_grad():
            baseline_output = baseline_model(inputs)
            optimized_output = solution_model(inputs)
            max_diff_avg += torch.max(torch.abs(optimized_output - baseline_output))
    max_diff_avg /= n_correctness_trials
    print(f"max float diff between values of baseline and optimized model: {max_diff_avg}")

    # measure performance
    inputs = get_inputs(batch_size=batch_size, seq_len=seq_len, n_embd=n_embd, device="cuda")
    t_avg_baseline = bench(baseline_model, inputs, n_warmup, n_rep)
    print(f"baseline time: {t_avg_baseline:.2f}ms")
    t_avg_optimized = bench(solution_model, inputs, n_warmup, n_rep)
    print(f"optimized time: {t_avg_optimized:.2f}ms")
    print(f"speedup: {t_avg_baseline / t_avg_optimized:.2f}x")



import torch
import torch.nn as nn
import torch.nn.functional as F
import math


class Model(nn.Module):
    """
    A vanilla multi-head masked self-attention layer with a projection at the end.
    """

    def __init__(self, n_embd, n_head, attn_pdrop, resid_pdrop, max_seqlen):
        super().__init__()
        assert n_embd % n_head == 0
        # key, query, value projections for all heads, but in a batch
        self.c_attn = nn.Linear(n_embd, 3 * n_embd)
        # output projection
        self.c_proj = nn.Linear(n_embd, n_embd)
        # regularization
        self.attn_dropout = nn.Dropout(attn_pdrop)
        self.resid_dropout = nn.Dropout(resid_pdrop)
        # causal mask to ensure that attention is only applied to the left in the input sequence
        self.register_buffer("bias", torch.tril(torch.ones(max_seqlen, max_seqlen)).view(1, 1, max_seqlen, max_seqlen))
        self.n_head = n_head
        self.n_embd = n_embd

    def forward(self, x):
        B, T, C = x.size()  # batch size, sequence length, embedding dimensionality (n_embd)
        # calculate query, key, values for all heads in batch and move head forward to be the batch dim
        q, k, v = self.c_attn(x).split(self.n_embd, dim=2)
        k = k.view(B, T, self.n_head, C // self.n_head).transpose(1, 2)  # (B, nh, T, hs)
        q = q.view(B, T, self.n_head, C // self.n_head).transpose(1, 2)  # (B, nh, T, hs)
        v = v.view(B, T, self.n_head, C // self.n_head).transpose(1, 2)  # (B, nh, T, hs)

        # causal self-attention; Self-attend: (B, nh, T, hs) x (B, nh, hs, T) -> (B, nh, T, T)
        att = (q @ k.transpose(-2, -1)) * (1.0 / math.sqrt(k.size(-1)))
        att = att.masked_fill(self.bias[:, :, :T, :T] == 0, float("-inf"))
        att = F.softmax(att, dim=-1)
        att = self.attn_dropout(att)
        y = att @ v  # (B, nh, T, T) x (B, nh, T, hs) -> (B, nh, T, hs)
        y = y.transpose(1, 2).contiguous().view(B, T, C)  # re-assemble all head outputs side by side
        # output projection
        y = self.resid_dropout(self.c_proj(y))
        return y



from rich.tree import Tree
from rich.table import Table
from rich.progress import BarColumn, Progress, TextColumn
from rich.layout import Layout
from rich.panel import Panel
from rich.syntax import Syntax
from typing import Dict, List, Optional, Union, Tuple
from .utils import format_number


class SummaryPanel:
    """Holds a summary of the optimization session."""

    def __init__(self, maximize: bool, metric_name: str, total_steps: int, model: str, session_id: str = None):
        self.maximize = maximize
        self.metric_name = metric_name
        self.goal = ("Maximizing" if self.maximize else "Minimizing") + f" {self.metric_name}..."
        self.total_input_tokens = 0
        self.total_output_tokens = 0
        self.total_steps = total_steps
        self.model = model
        self.session_id = session_id or "N/A"
        self.progress = Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(bar_width=20),
            TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
            TextColumn("•"),
            TextColumn("[bold]{task.completed}/{task.total} Steps"),
            expand=False,
        )
        self.task_id = self.progress.add_task("", total=total_steps)

    def set_step(self, step: int):
        """Set the current step."""
        self.progress.update(self.task_id, completed=step)

    def update_token_counts(self, usage: Dict[str, int]):
        """Update token counts from usage data."""
        if not isinstance(usage, dict) or "input_tokens" not in usage or "output_tokens" not in usage:
            raise ValueError("Invalid token usage response from API.")
        self.total_input_tokens += usage["input_tokens"]
        self.total_output_tokens += usage["output_tokens"]

    def get_display(self, final_message: Optional[str] = None) -> Panel:
        """Create a summary panel with the relevant information."""
        layout = Layout(name="summary")
        summary_table = Table(show_header=False, box=None, padding=(0, 1))
        # Goal
        if final_message is not None:
            summary_table.add_row(f"[bold cyan]Result:[/] {final_message}")
        else:
            summary_table.add_row(f"[bold cyan]Goal:[/] {self.goal}")
        summary_table.add_row("")
        # Model used
        summary_table.add_row(f"[bold cyan]Model:[/] {self.model}")
        summary_table.add_row("")
        # Log directory
        runs_dir = f".runs/{self.session_id}"
        summary_table.add_row(f"[bold cyan]Logs:[/] [blue underline]{runs_dir}[/]")
        summary_table.add_row("")
        # Token counts
        summary_table.add_row(
            f"[bold cyan]Tokens:[/] ↑[yellow]{format_number(self.total_input_tokens)}[/] ↓[yellow]{format_number(self.total_output_tokens)}[/] = [green]{format_number(self.total_input_tokens + self.total_output_tokens)}[/]"
        )
        summary_table.add_row("")
        # Progress bar
        summary_table.add_row(self.progress)

        # Update layout
        layout.update(summary_table)

        return Panel(layout, title="[bold]📊 Summary", border_style="magenta", expand=True, padding=(0, 1))


class PlanPanel:
    """Displays the optimization plan with truncation for long plans."""

    def __init__(self):
        self.plan = ""

    def update(self, plan: str):
        """Update the plan text."""
        self.plan = plan

    def clear(self):
        """Clear the plan text."""
        self.plan = ""

    def get_display(self) -> Panel:
        """Create a panel displaying the plan with truncation if needed."""
        return Panel(self.plan, title="[bold]📝 Thinking...", border_style="cyan", expand=True, padding=(0, 1))


class Node:
    """Represents a node in the solution tree."""

    def __init__(
        self, id: str, parent_id: Union[str, None], code: Union[str, None], metric: Union[float, None], is_buggy: bool
    ):
        self.id = id
        self.parent_id = parent_id
        self.children: List["Node"] = []
        self.code = code
        self.metric = metric
        self.is_buggy = is_buggy
        self.evaluated = True


class MetricTree:
    """Manages the tree structure of optimization solutions."""

    def __init__(self, maximize: bool):
        self.nodes: Dict[str, Node] = {}
        self.maximize = maximize

    def clear(self):
        """Clear the tree."""
        self.nodes = {}

    def add_node(self, node: Node):
        """Add a node to the tree."""
        # Add the node to the tree
        self.nodes[node.id] = node

        # Add node to node's parent's children
        if node.parent_id is not None:
            if node.parent_id not in self.nodes:
                raise ValueError("Could not construct tree: parent node not found.")
            self.nodes[node.parent_id].children.append(node)

    def get_draft_nodes(self) -> List[Node]:
        """Get all draft nodes from the tree."""
        return [node for node in self.nodes.values() if node.parent_id is None]

    def get_best_node(self) -> Optional[Node]:
        """Get the best node from the tree."""
        measured_nodes = [
            node
            for node in self.nodes.values()
            if node.evaluated  # evaluated
            and not node.is_buggy  # not buggy
            and node.metric is not None  # has metric
        ]
        if len(measured_nodes) == 0:
            return None
        if self.maximize:
            return max(measured_nodes, key=lambda node: node.metric)
        else:
            return min(measured_nodes, key=lambda node: node.metric)


class MetricTreePanel:
    """Displays the solution tree with depth limiting."""

    def __init__(self, maximize: bool):
        self.metric_tree = MetricTree(maximize=maximize)

    def build_metric_tree(self, nodes: List[dict]):
        """Build the tree from the list of nodes."""
        # First clear then tree
        self.metric_tree.clear()

        # Then sort the nodes by step number
        nodes.sort(key=lambda x: x["step"])

        # Finally build the new tree
        for node in nodes:
            self.metric_tree.add_node(
                Node(
                    id=node["solution_id"],
                    parent_id=node["parent_id"],
                    code=node["code"],
                    metric=node["metric_value"],
                    is_buggy=node["is_buggy"],
                )
            )

    def set_unevaluated_node(self, node_id: str):
        """Set the unevaluated node."""
        self.metric_tree.nodes[node_id].evaluated = False

    def _build_rich_tree(self) -> Tree:
        """Get a Rich Tree representation of the solution tree using a DFS like traversal."""
        if len(self.metric_tree.nodes) == 0:
            return Tree("[bold green]Building first solution...")

        best_node = self.metric_tree.get_best_node()

        def append_rec(node: Node, tree: Tree):
            if not node.evaluated:
                # not evaluated
                color = "yellow"
                style = None
                text = "evaluating..."
            elif node.is_buggy:
                # buggy node
                color = "red"
                style = None
                text = "bug"
            else:
                # evaluated non-buggy node
                if node.id == best_node.id:
                    # best node
                    color = "green"
                    style = "bold"
                    text = f"{node.metric:.3f} 🏆"
                elif node.metric is None:
                    # metric not extracted from evaluated solution
                    color = "yellow"
                    style = None
                    text = "N/A"
                else:
                    # evaluated node with metric
                    color = "green"
                    style = None
                    text = f"{node.metric:.3f}"

            s = f"[{f'{style} ' if style is not None else ''}{color}]● {text}"
            subtree = tree.add(s)
            for child in node.children:
                append_rec(child, subtree)

        tree = Tree("🌳")
        for n in self.metric_tree.get_draft_nodes():
            append_rec(n, tree)

        return tree

    def get_display(self) -> Panel:
        """Get a panel displaying the solution tree."""
        # Make sure the metric tree is built before calling build_rich_tree
        return Panel(
            self._build_rich_tree(), title="[bold]🔎 Exploring Solutions...", border_style="green", expand=True, padding=(0, 1)
        )


class EvaluationOutputPanel:
    """Displays evaluation output with truncation for long outputs."""

    def __init__(self):
        self.output = ""

    def update(self, output: str) -> None:
        """Update the evaluation output."""
        self.output = output

    def clear(self) -> None:
        """Clear the evaluation output."""
        self.output = ""

    def get_display(self) -> Panel:
        """Create a panel displaying the evaluation output with truncation if needed."""
        return Panel(self.output, title="[bold]📋 Evaluation Output", border_style="blue", expand=True, padding=(0, 1))


class SolutionPanels:
    """Displays the current and best solutions side by side."""

    def __init__(self, metric_name: str):
        # Current solution
        self.current_node = None
        # Best solution
        self.best_node = None
        # Metric name
        self.metric_name = metric_name.capitalize()

    def update(self, current_node: Union[Node, None], best_node: Union[Node, None]):
        """Update the current and best solutions."""
        # Update current solution
        self.current_node = current_node
        # Update best solution
        self.best_node = best_node

    def get_display(self, current_step: int) -> Tuple[Panel, Panel]:
        """Return the current and best solutions as panels."""
        current_code = self.current_node.code if self.current_node is not None else ""
        best_code = self.best_node.code if self.best_node is not None else ""
        best_score = self.best_node.metric if self.best_node is not None else None

        # Current solution (without score)
        current_title = f"[bold]💡 Current Solution (Step {current_step})"
        current_panel = Panel(
            Syntax(str(current_code), "python", theme="monokai", line_numbers=True, word_wrap=False),
            title=current_title,
            border_style="yellow",
            expand=True,
            padding=(0, 1),
        )

        # Best solution
        best_title = f"[bold]🏆 Best Solution ([green]{self.metric_name}: {f'{best_score:.4f}' if best_score is not None else 'N/A'}[/])"
        best_panel = Panel(
            Syntax(str(best_code), "python", theme="monokai", line_numbers=True, word_wrap=False),
            title=best_title,
            border_style="green",
            expand=True,
            padding=(0, 1),
        )

        return current_panel, best_panel


def create_optimization_layout() -> Layout:
    """Create the main layout for the CLI."""
    layout = Layout()

    # First split into top, middle, and bottom sections
    layout.split_column(
        Layout(name="top_section", ratio=3), Layout(name="middle_section", ratio=4), Layout(name="eval_output", ratio=2)
    )

    # Split the top section into left and right
    layout["top_section"].split_row(Layout(name="left_panels", ratio=1), Layout(name="tree", ratio=1))

    # Split the left panels into summary and thinking
    layout["left_panels"].split_column(Layout(name="summary", ratio=2), Layout(name="plan", ratio=1))

    # Split the middle section into left and right
    layout["middle_section"].split_row(Layout(name="current_solution", ratio=1), Layout(name="best_solution", ratio=1))

    return layout


def create_end_optimization_layout() -> Layout:
    """Create the final layout after optimization is complete."""
    layout = Layout()

    # Create a top section for summary
    layout.split_column(Layout(name="summary", ratio=1), Layout(name="bottom_section", ratio=3))

    # Split the bottom section into left (best solution) and right ( tree)
    layout["bottom_section"].split_row(Layout(name="best_solution", ratio=1), Layout(name="tree", ratio=1))

    return layout



from typing import Any, Dict, List, Tuple, Union
import json
import os
import time
import subprocess
from rich.layout import Layout
from rich.live import Live
from rich.panel import Panel
import pathlib


# Env/arg helper functions
def read_api_keys_from_env() -> Dict[str, Any]:
    """Read API keys from environment variables."""
    keys = {}
    keys_to_check = ["OPENAI_API_KEY", "ANTHROPIC_API_KEY", "GEMINI_API_KEY"]
    for key in keys_to_check:
        value = os.getenv(key)
        if value is not None and len(value) > 0:
            keys[key] = value
    return keys


def read_additional_instructions(additional_instructions: str | None) -> str | None:
    """Read additional instructions from a file path string or return the string itself."""
    if additional_instructions is None:
        return None

    # Try interpreting as a path first
    potential_path = pathlib.Path(additional_instructions)
    try:
        if potential_path.exists() and potential_path.is_file():
            return read_from_path(potential_path, is_json=False)  # type: ignore # read_from_path returns str when is_json=False
        else:
            # If it's not a valid file path, return the string itself
            return additional_instructions
    except OSError:
        # If the path can't be read, return the string itself
        return additional_instructions


# File helper functions
def read_from_path(fp: pathlib.Path, is_json: bool = False) -> Union[str, Dict[str, Any]]:
    """Read content from a file path, optionally parsing as JSON."""
    with fp.open("r", encoding="utf-8") as f:
        if is_json:
            return json.load(f)
        return f.read()


def write_to_path(fp: pathlib.Path, content: Union[str, Dict[str, Any]], is_json: bool = False) -> None:
    """Write content to a file path, optionally as JSON."""
    with fp.open("w", encoding="utf-8") as f:
        if is_json:
            json.dump(content, f, indent=4)
        elif isinstance(content, str):
            f.write(content)
        else:
            raise TypeError("Content must be str or Dict[str, Any]")


# Visualization helper functions
def format_number(n: Union[int, float]) -> str:
    """Format large numbers with K, M, B, T suffixes for better readability."""
    if n >= 1e12:
        return f"{n / 1e12:.1f}T"
    elif n >= 1e9:
        return f"{n / 1e9:.1f}B"
    elif n >= 1e6:
        return f"{n / 1e6:.1f}M"
    elif n >= 1e3:
        return f"{n / 1e3:.1f}K"
    # Handle potential floats that don't need suffix but might need formatting
    if isinstance(n, float):
        # Format floats nicely, avoid excessive precision unless needed
        return f"{n:.4g}"  # Use general format, up to 4 significant digits
    return str(n)


def smooth_update(
    live: Live, layout: Layout, sections_to_update: List[Tuple[str, Panel]], transition_delay: float = 0.05
) -> None:
    """
    Update sections of the layout with a small delay between each update for a smoother transition effect.

    Args:
        live: The Live display instance
        layout: The Layout to update
        sections_to_update: List of (section_name, content) tuples to update
        transition_delay: Delay in seconds between updates (default: 0.05)
    """
    for section, content in sections_to_update:
        layout[section].update(content)
        live.refresh()
        time.sleep(transition_delay)


# Other helper functions
def run_evaluation(eval_command: str) -> str:
    """Run the evaluation command on the code and return the output."""

    # Run the eval command as is
    result = subprocess.run(eval_command, shell=True, capture_output=True, text=True, check=False)

    # Combine stdout and stderr for complete output
    output = result.stderr if result.stderr else ""
    if result.stdout:
        if len(output) > 0:
            output += "\n"
        output += result.stdout
    return output
